# from smbgym import Env
"""Registration code of Gym environments in this package."""
from .env import Env
# from .smb_random_stages_env import SuperMarioBrosRandomStagesEnv
from ._registration import make


# define the outward facing API of this package
__all__ = [
    make.__name__,
    Env.__name__,
    # SuperMarioBrosRandomStagesEnv.__name__,
]


from py4j.java_gateway import JavaGateway
import py4j
import numpy as np
import sys
import os

class Bridge:
	"""
	A bridge between Python and the Java Mario Environment
	"""

	def __init__(self, visuals=False) -> None:
		self.visuals = visuals
		
		self._connect()
	
	def _connect(self) -> None:
		level_path = os.path.join(os.path.dirname(__file__), "./bin/ap.jar")
		self.gateway = JavaGateway.launch_gateway(classpath=level_path, die_on_exit=True, redirect_stdout=sys.stdout, redirect_stderr=sys.stderr)
		self.root = self.gateway.jvm.PlayLevel()

	def set_level(self, path) -> str:
		py4j.java_gateway.set_field(self.root, 'level', path)
		return path
	
	def initalize(self) -> None:
		if self.visuals == "human":
			self.root.initializeWithGraphics()
		else:
			self.root.initializeHeadless()

		self.agent = py4j.java_gateway.get_field(self.root, 'agent')
		self.game = py4j.java_gateway.get_field(self.root, 'game')

		self.world = py4j.java_gateway.get_field(self.game, "world")
		self.mario = py4j.java_gateway.get_field(self.world, "mario")

		self.game.step()

	
	def reset(self) -> None:
		self.initalize()

	def step(self, action) -> None:
		self.register_inputs(action)
		self.game.step()

	def _get_coins(self):
		"""
		Get the number of coins collected
		"""
		return py4j.java_gateway.get_field(self.world, "coins")

	def _get_lives(self):
		"""
		Get the number of remaining lives
		"""
		lives = py4j.java_gateway.get_field(self.world, "lives")
		if self._get_game_status() == "LOSE":
			lives -= 1
		return lives

	def _get_XY(self):
		"""
		Get the X and Y pos of Mario
		"""
		x = py4j.java_gateway.get_field(self.mario, "x")
		y = py4j.java_gateway.get_field(self.mario, "y")
		return (x, y)
	
	def _get_game_status(self):
		"""
		Get the status of the game (RUNNING, WIN, LOSE, TIME_OUT)
		"""
		return str(py4j.java_gateway.get_field(self.world, "gameStatus"))
	
	def _flag_get(self):
		"""
		Returns a Boolean on wheather the flag has been touched
		"""
		status = self._get_game_status()
		if status == "WIN":
			return True
		return False
	
	def _get_mario_status(self):
		"""
		Returns the status of Mario (fireball, big, small)
		"""
		large = py4j.java_gateway.get_field(self.mario, "isLarge")
		fire = py4j.java_gateway.get_field(self.mario, "isFire")
		if fire:
			return "fireball"
		elif large:
			return "big"
		else:
			return "small"
	
	def _get_time (self):
		return py4j.java_gateway.get_field(self.world, "currentTimer") / 1000

	def get_observation(self):
		xy = self._get_XY()
		x = xy[0]
		y = xy[1]
		return self.world.getMergedObservation(x, y)

	def shutdown(self) -> None:
		self.gateway.shutdown()
	
	def register_inputs(self, action):
		self.agent.clear()

		if action == 0:
			pass
		elif action == 1:
			self.agent.right()
		elif action == 2:
			self.agent.right()
			self.agent.jump()
		elif action == 3:
			self.agent.right()
			self.agent.speed()
		elif action == 4:
			self.agent.right()
			self.agent.jump()
			self.agent.speed()
		elif action == 5:
			self.agent.jump()
		elif action == 6:
			self.agent.left()
		elif action == 7:
			self.agent.left()
			self.agent.jump()
		elif action == 8:
			self.agent.left()
			self.agent.speed()
		elif action == 9:
			self.agent.left()
			self.agent.jump()
			self.agent.speed()
		elif action == 10:
			self.agent.down()
		elif action == 11:
			# self.agent.up()
			pass
	
	def get_human_observation(self):
		"""
		Returns an observation 
		"""
		screen = self.get_observation()
		screen = np.array(screen)
		return np.flip(np.rot90(screen, 1, (0,1)), 0)
	
	def get_info(self):
		"""
		Returns a Dictionary of information from the environment
		"""
		xy = self._get_XY()
		return dict(
			coins = self._get_coins(),
			flag_get = self._flag_get(),
			life = self._get_lives(),
			score = 0,
			stage = None,
			status = self._get_mario_status(),
			time = self._get_time(),
			world = None,
			x = xy[0],
			y = xy[1]
		)


from smbgym.bridge import Bridge
import numpy as np
from smbgym.helper.encode import encode
import gym
from os import path
import random

import time

class Env(gym.Env):
	"""
	Another environment for playing Super Mario Bros with OpenAI Gym

	Based on Gym Super Mario Bros (credit: Kautenja https://github.com/Kautenja) and Mario AI Framework (credit: amidos2006 https://github.com/amidos2006)
	You guys are awesome and this would not exist without you

	"""

	metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 30}


	def __init__(self, level_mode="random", render_mode="human") -> None:
		self.bridge = Bridge(render_mode)

		self.level_mode = level_mode
		self._set_level()

		self.bridge.initalize()

		info = self._get_info()
		self.mario_inital_x = info["x"]

		self.action_space = gym.spaces.Discrete(12)

		assert render_mode is None or render_mode in self.metadata["render_modes"]
		self.render_mode = render_mode

	def _set_level(self):
		dir_path = path.join(path.dirname(__file__), "./original")
		if self.level_mode == "random":
			num = random.randint(1, 15)
			level = path.join(dir_path, "./" + "lvl-" + str(num) + ".txt")
			print(level)
			self.bridge.set_level(level)
		else:
			level = path.join(dir_path, "./" + self.level_mode)
			self.bridge.set_level(level)
	
	def _encode_state(self, obs) -> list:
		indexer = np.array([encode.get(i, [0,0,0]) for i in range(0, 255)])
		return indexer[(obs)]
	
	def _get_observation(self) -> list:
		obs = self.bridge.get_observation()
		obs = self._encode_state(obs)
		return np.array(obs, dtype=np.uint8)
	
	def _get_info(self) -> dict:
		return self.bridge.get_info()
	
	def _calculate_reward(self, info) -> float:
		v = info["x"] - self.mario_inital_x
		t = -0.1
		d = 0
		if info["life"] == 0:
			d = -15
		
		r = v + t + d
		
		if -15 > r:
			r = -15
		if 15 < r:
			r = 15

		return r

	def close(self) -> None:
		if self.render_mode == "human":
			self.bridge.root.closeWindow()
		self.bridge.gateway.close()


	def step(self, action) -> tuple[list, float, bool, bool, dict]:

		self.bridge.step(action)

		obs = self._get_observation()
		info = self._get_info()
		reward = self._calculate_reward(info)
		if (info["flag_get"] == True or info["life"] == 0) or (info["time"] == 0):
			terminated = True
		else:
			terminated = False

		return (
			obs,
			reward,
			terminated,
			False,
			info
		)

	def reset(self) -> tuple[list, float, bool, bool, dict]:
		self._set_level()
		self.bridge.reset()

		return (
			self._get_observation(),
			self._get_info()
		)

if __name__ == "__main__":
	env = Env()
	while True:
		obs, reward, term, trunc, info = env.step([1, 0, 0, 0, 0])
		print(obs.shape)
		print(info)
		if info['life'] == 0:
			info = env.reset()
			break


"""The setup script for installing the package."""
from setuptools import setup, find_packages


# # read the contents of the README
# with open('README.md') as README_md:
#     README = README_md.read()


setup(
    name='smbgym',
    version='0.0.2',
    description='Super Mario Bros. for OpenAI Gym',
    # long_description=README,
    # long_description_content_type='text/markdown',
    keywords=' '.join([
        'OpenAI-Gym'
        'Super-Mario-Bros',
        'Reinforcement-Learning-Environment'
    ]),
    classifiers=[
        'Development Status :: 1 - Active Development',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: Free For Educational Use',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Games/Entertainment :: Side-Scrolling/Arcade Games',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    url='https://github.com/Mike968/smb-gym',
    author='Michael Mehall',
    author_email='michaelhmehall@gmail.com',
    license='Proprietary',
    packages=find_packages(exclude=['tests', '*.tests', '*.tests.*']),
    # package_data={ 'smb-gym': ['_roms/*.nes'] },
    install_requires=['py4j>=0.10.0'],
    include_package_data=True
    # entry_points={
    #     'console_scripts': [
    #         'gym_super_mario_bros = gym_super_mario_bros._app.cli:main',
    #     ],
    # },
)